<?php

namespace App\Http\Controllers\User;

use App\Constants\Status;
use App\Http\Controllers\Controller;
use App\Lib\CurrencyExchanger;
use App\Lib\FormProcessor;
use App\Lib\Intended;
use App\Models\Admin;
use App\Models\AdminNotification;
use App\Models\Currency;
use App\Models\Deposit;
use App\Models\Exchange;
use App\Models\GatewayCurrency;
use App\Models\RateAlert;
use Barryvdh\DomPDF\Facade\Pdf;
use Exception;
use Illuminate\Http\Request;

class ExchangeController extends Controller {
    public function exchange(Request $request) {
        $currencyExchanger = new CurrencyExchanger();
        $message           = $currencyExchanger->currencyExchanger($request);

        if ($message['status'] == 'error') {
            return back()->withNotify($message['notify'])->withInput();
        }

        if (!auth()->user()) {
            session()->put('exchange_data', $currencyExchanger);
            Intended::identifyRoute();
            return to_route('user.login');
        }

        $currencyExchanger->createExchange();
        return to_route('user.exchange.preview');
    }

    public function preview() {
        if (!session()->has('EXCHANGE_TRACK')) {
            $notify[] = ['error', "Invalid session"];
            return to_route('home')->withNotify($notify);
        }
        $pageTitle = 'Exchange Preview';
        $exchange  = Exchange::where('exchange_id', session('EXCHANGE_TRACK'))->with('receivedCurrency.userDetailsData')->firstOrFail();
        return view('Template::user.exchange.preview', compact('pageTitle', 'exchange'));
    }

    public function confirm(Request $request) {
        if (!session()->has('EXCHANGE_TRACK')) {
            $notify[] = ['error', "Invalid session"];
            return to_route('home')->withNotify($notify);
        }
        $validation = [
            'wallet_id' => 'required',
        ];

        $exchange         = Exchange::where('exchange_id', session()->get('EXCHANGE_TRACK'))->firstOrFail();
        $userRequiredData = @$exchange->receivedCurrency->userDetailsData->form_data ?? [];

        $formProcessor  = new FormProcessor();
        $validationRule = $formProcessor->valueValidation($userRequiredData);
        $validationRule = array_merge($validationRule, $validation);
        $request->validate($validationRule);

        $userData            = $formProcessor->processFormData($request, $userRequiredData);
        $exchange->user_data = $userData ?? null;
        $exchange->wallet_id = $request->wallet_id;
        $exchange->save();

        //=====automatic payment

        if ($exchange->sendCurrency->gateway_id != 0) {
            return $this->createDeposit($exchange);
        }
        return to_route('user.exchange.manual');
    }

    public function manual() {
        if (!session()->has('EXCHANGE_TRACK')) {
            $notify[] = ['error', "Something went the wrong with exchange processing"];
            return to_route('home')->withNotify($notify);
        }
        $exchange  = Exchange::where('exchange_id', session()->get('EXCHANGE_TRACK'))->firstOrFail();
        $pageTitle = "Transaction Proof";
        return view('Template::user.exchange.manual', compact('pageTitle', 'exchange'));
    }

    public function manualConfirm(Request $request) {
        if (!session()->has('EXCHANGE_TRACK')) {
            $notify[] = ['error', "Something went the wrong with exchange processing"];
            return to_route('home')->withNotify($notify);
        }

        $exchange              = Exchange::where('exchange_id', session()->get('EXCHANGE_TRACK'))->firstOrFail();
        $transactionProvedData = @$exchange->sendCurrency->transactionProvedData->form_data ?? [];

        $formProcessor  = new FormProcessor();
        $validationRule = $formProcessor->valueValidation($transactionProvedData);
        $request->validate($validationRule);
        $provedData = $formProcessor->processFormData($request, $transactionProvedData);

        $exchange->transaction_proof_data = $provedData ?? null;
        $exchange->status                 = Status::EXCHANGE_PENDING;
        $exchange->save();

        $comment = 'send ' . getAmount($exchange->get_amount) . ' by ' . @$exchange->sendCurrency->name;

        $adminNotification            = new AdminNotification();
        $adminNotification->user_id   = $exchange->user_id;
        $adminNotification->title     = $comment;
        $adminNotification->click_url = urlPath('admin.exchange.details', $exchange->id);
        $adminNotification->save();

        $admin = Admin::first();

        if ($admin && gs('admin_email_notification') == Status::YES && isset($exchange)) {
            $user = (object) [
                'username' => $admin->username,
                'email'    => $admin->email,
                'fullname' => $admin->name,
            ];

            $shortCodes = [
                'username'         => $admin->username,
                'exchange_id'      => $exchange->exchange_id ?? '',
                'amount'           => getAmount($exchange->sending_amount ?? 0),
                'send_currency'    => optional($exchange->sendCurrency)->name,
                'receive_currency' => optional($exchange->receivedCurrency)->name,
                'exchange_link'    => urlPath('admin.exchange.details', $exchange->id),
            ];

            notify($user, 'EXCHANGE_APPROVAL_REQUIRED', $shortCodes, ['email']);
        }

        session()->forget('EXCHANGE_TRACK');

        $notify[] = ['success', 'Admin will review your request'];
        return to_route('user.exchange.details', $exchange->exchange_id)->withNotify($notify);
    }

    public function list($scope = 'list') {
        try {
            $exchanges = Exchange::$scope()->where('user_id', auth()->id())->with(['sendCurrency', 'receivedCurrency'])->desc()->paginate(getPaginate());
            $pageTitle = formateScope($scope) . " Exchange";
        } catch (Exception $ex) {
            $notify[] = ['error', 'Invalid URL.'];
            return back()->withNotify($notify);
        }
        return view('Template::user.exchange.list', compact('pageTitle', 'exchanges'));
    }

    public function details($trx) {
        $exchange = Exchange::where('user_id', auth()->id())->with('deposit', function ($deposit) {
            $deposit->where('status', Status::PAYMENT_INITIATE);
        })->where('exchange_id', $trx)->firstOrFail();
        $pageTitle = 'Exchange Details';
        return view('Template::user.exchange.details', compact('pageTitle', 'exchange'));
    }

    protected function validation($request) {
        $request->validate([
            'sending_amount'     => 'required|numeric|gt:0',
            'receiving_amount'   => 'required|numeric|gt:0',
            'sending_currency'   => 'required|integer',
            'receiving_currency' => 'required|integer|different:sending_currency',
        ]);
    }

    public function invoice($exchangeId, $type) {
        $types = ['print', 'download'];

        if (!in_array($type, $types)) {
            $notify[] = ['error', "Invalid URL."];
            return to_route('user.exchange.list', 'list')->withNotify($notify);
        }
        if ($type == 'print') {
            $pageTitle = "Print Exchange";
            $action    = 'stream';
        } else {
            $pageTitle = "Download Exchange";
            $action    = 'download';
        }

        $user     = auth()->user();
        $exchange = Exchange::where('status', '!=', Status::EXCHANGE_INITIAL)
            ->where('exchange_id', $exchangeId)
            ->where('user_id', $user->id)->firstOrFail();

        $pdf      = PDF::loadView('partials.pdf', compact('pageTitle', 'user', 'exchange'));
        $fileName = $exchange->exchange_id . '_' . time();

        return $pdf->$action($fileName . '.pdf');
    }

    public function complete($id) {
        $exchange = Exchange::where('user_id', auth()->id())->where('id', $id)->where('status', Status::EXCHANGE_INITIAL)->firstOrFail();

        session()->put('EXCHANGE_TRACK', $exchange->exchange_id);
        if (!$exchange->wallet_id) {
            return to_route('user.exchange.preview');
        }

        if ($exchange->sendCurrency->gateway_id && !$exchange->automatic_payment_status) {
            if (!$exchange->deposit) {
                return $this->createDeposit($exchange);
            }
            session()->put('Track', $exchange->exchange_id);
            return to_route('user.deposit.confirm');
        }

        if (!$exchange->sendCurrency->gateway_id && !$exchange->transaction_proof_data) {
            return to_route('user.exchange.manual');
        }
        return back();
    }

    private function createDeposit($exchange) {
        $curSymbol = $exchange->sendCurrency->cur_sym;
        $code      = $exchange->sendCurrency->gatewayCurrency->code;
        $gateway   = GatewayCurrency::where('method_code', $code)->where('currency', $curSymbol)->first();

        if (!$gateway) {
            $notify[] = ['error', "Something went the wrong with exchange processing"];
            return back()->withNotify($notify);
        }
        $amount = $exchange->sending_amount + $exchange->sending_charge;

        $deposit                  = new Deposit();
        $deposit->user_id         = auth()->id();
        $deposit->method_code     = $code;
        $deposit->method_currency = strtoupper($curSymbol);
        $deposit->amount          = $amount;
        $deposit->charge          = $exchange->sending_charge;
        $deposit->rate            = $exchange->buy_rate;
        $deposit->final_amount    = $amount;
        $deposit->btc_amount      = 0;
        $deposit->btc_wallet      = "";
        $deposit->trx             = $exchange->exchange_id;
        $deposit->try             = 0;
        $deposit->success_url     = urlPath('user.exchange.list');
        $deposit->failed_url      = urlPath('user.exchange.list');
        $deposit->status          = 0;
        $deposit->exchange_id     = $exchange->id;
        $deposit->save();

        session()->put('Track', $deposit->trx);
        return to_route('user.deposit.confirm');
    }

    public function getExchangeRate(Request $request) {
        $request->validate([
            'from_currency' => 'required|integer|exists:currencies,id',
            'to_currency'   => 'required|integer|exists:currencies,id|different:from_currency',
            'target_rate'   => 'required|numeric|min:0.1',
            'alert_email'   => 'required|email',
            'expire_time'   => 'required',
        ]);

        $isExistsAlert = RateAlert::where('from_currency_id', $request->from_currency)->where('to_currency_id', $request->to_currency)->where('alert_email', $request->alert_email)->where('status', Status::ALERT_PENDING)->exists();
        if ($isExistsAlert) {
            $notify[] = ['error', "Rate alert for this currency pair already exists"];
            return back()->withNotify($notify);
        }

        $expireDuration = $request->expire_time;
        $expireTime     = now();

        switch ($expireDuration) {
        case '6':
            $expireTime = $expireTime->addHours(6);
            break;
        case '12':
            $expireTime = $expireTime->addHours(12);
            break;
        case '24':
            $expireTime = $expireTime->addHours(24);
            break;
        case 'week':
            $expireTime = $expireTime->addWeek();
            break;
        case 'month':
            $expireTime = $expireTime->addMonth();
            break;
        case '3-months':
            $expireTime = $expireTime->addMonths(3);
            break;
        }

        $rateAlert                   = new RateAlert();
        $rateAlert->from_currency_id = $request->from_currency;
        $rateAlert->to_currency_id   = $request->to_currency;
        $rateAlert->target_rate      = $request->target_rate;
        $rateAlert->alert_email      = $request->alert_email;
        $rateAlert->expire_time      = $expireTime;
        $rateAlert->save();

        $notify[] = ['success', "Notification alert has been saved successfully"];
        return back()->withNotify($notify);
    }

    public function bestRates(Request $request) {
        $sendingCurrencyId   = $request->input('sending_currency');
        $receivingCurrencyId = $request->input('receiving_currency');

        $sendingCurrency = Currency::where('id', $sendingCurrencyId)
            ->where('available_for_sell', Status::YES)
            ->where('sell_at', '>', 0)
            ->select(['id', 'name as sending_currency', 'sell_at', 'buy_at', 'cur_sym as send_currency_symbol'])
            ->first();

        if (!$sendingCurrency) {
            return response()->json(['rates' => []]);
        }

        $sendCurrencyBuyAt = $sendingCurrency->buy_at;

        $receivingCurrency = Currency::where('id', $receivingCurrencyId)->first();
        if (!$receivingCurrency) {
            return response()->json(['rates' => []]);
        }

        $receivingCurrencySymbol = $receivingCurrency->cur_sym;

        $currencies = Currency::where('available_for_sell', Status::YES)
            ->where('id', '!=', $sendingCurrencyId)
            ->where('cur_sym', $receivingCurrencySymbol)
            ->select([
                'id',
                'name as receiving_currency',
                'sell_at',
                'cur_sym as receive_currency_symbol',
            ])
            ->get();

        $calculatedRates = $currencies->map(function ($currency) use ($sendingCurrency, $sendCurrencyBuyAt) {
            $exchangeRate = ($sendCurrencyBuyAt / $currency->sell_at) * 1;

            return [
                'sending_currency'        => $sendingCurrency->sending_currency,
                'receiving_currency'      => $currency->receiving_currency,
                'rate'                    => number_format($exchangeRate, 2),
                'currency_id'             => $currency->id,
                'receive_currency_symbol' => $currency->receive_currency_symbol,
                'send_currency_symbol'    => $sendingCurrency->send_currency_symbol,
            ];
        });

        $sortedRates = $calculatedRates->sortByDesc('rate')->values();

        return response()->json(['rates' => $sortedRates]);
    }
}
